//   FigGraphics.java --- FIG graphics for JAVA
//   This is a part of Fungus.

//   Copyright (C) 2001 Valvassori Mose <djedi@altern.org>

//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.

//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package fungus.core.utils;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;

/**
    FIG Graphics<br>
    This is a graphics subclass that produce a FIG 3.2.

    <p>Usage:<pre>
import fungus.core.util.FigGraphics;
     ...
     Graphics fig = new FigGraphics();

     fig.setColor(Color.red);
     fig.fillRect(10, 10, 30, 30);
</pre>

<p><ul>
  <li> 1 pixel = 0.02222 mm
  <li> 45 px = 1mm
  </ul>
<p>page layout:  A4 landscape

 * @author "Valvassori Mose" <djedi@altern.org>
 * @version $Id: FigGraphics.java,v 1.3 2001/01/24 22:09:22 djedi Exp $
 * @since 1.0
 * @see java.awt.Graphics
 */
public  class FigGraphics extends java.awt.Graphics
{
  public static final boolean DEBUG = true;
  
  protected final static int PAGEHEIGHT = 9449;
  protected final static int PAGEWIDTH = 13364;

  protected  int xOffset = 0;			  // Origin of the display
  protected  int yOffset = 0;

  protected PrintWriter os;


  public final static int cDefault = -1;
  public final static int cBlack = 0;
  public final static int cBlue = 1;
  public final static int cGreen = 2;
  public final static int cCyan = 3;
  public final static int cRed = 4;
  public final static int cMagenta = 5;
  public final static int cYellow = 6;
  public final static int cWhite = 7;



  /**
    The current color
  */
  protected Color clr = Color.black;
  protected int pen_color = cBlack;
  protected int fill_color = cWhite;

  /**
     The current font
  */
  protected Font font = new Font("Helvetica",Font.PLAIN,12);
  protected Rectangle clippingRect = new Rectangle(0,0,PAGEWIDTH,PAGEHEIGHT);

  protected int line_style = lSDefault;
  public final static int lSDefault = -1;
  public final static int lSSolid = 0;
  public final static int lSDashed = 1;
  public final static int lSDotted = 2;
  public final static int lSDash_dotted = 3;
  public final static int lSDash_double_dotted = 4;
  public final static int lSDash_triple_dotted = 5;

  public final static int afNotFilled = -1;
  public final static int afFullSaturation = 20;

  protected int depth = 50;
  protected int area_fill = afNotFilled;
  protected float style_val = (float)0.000;

  protected int join_style = jMiter;
  public final static int jMiter = 0;
  public final static int jBevel = 1;
  public final static int jRound = 2;

  protected int cap_style = csButt;
  public final static int csButt = 0;
  public final static int csRound = 1;
  public final static int csPrjecting = 2;

  protected int radius = -1;

  protected int forward_arrow = aOff;
  protected int backward_arrow = aOff;
  public final static int aOff = 0;
  public final static int aOn = 1;


  public FigGraphics()
  {
    this(new OutputStreamWriter(System.out), true);
  }

  public FigGraphics(Writer w)
  {
    this(w, true);
  }

  public FigGraphics(Writer w, boolean emitProlog)
  {
    setOutput(w);
    
    if (emitProlog)
      emitProlog();
  }

  public Graphics create()
  {
    throw new RuntimeException("create() not implemented");
  }

  /**
   * Change the Writer this context's output goes to.
   */

  public void setOutput(Writer w)
  {
    if (! (w instanceof PrintWriter))
      os = new PrintWriter(w, true);
    else
      os = (PrintWriter) w;    
  }


  private void preparePolyline(int sub_type){
    os.print("2 ");				  // polyline type
    os.print(sub_type+" ");
    os.print(line_style+" ");
    os.print("1 ");				  // tickness
    os.print(pen_color+" ");
    os.print(fill_color+" ");
    os.print(depth+" ");
    os.print("0 ");
    os.print(area_fill+" ");
    os.print(style_val+" ");
    os.print(join_style+" ");
    os.print(cap_style+" ");
    os.print(radius+" ");
    os.print(forward_arrow+" ");
    os.print(backward_arrow+" ");
  }

  private void prepareOval(int sub_type){
    os.print("1 ");
    os.print(sub_type+" ");
    os.print(line_style+" ");
    os.print("1 ");				  // tickness
    os.print(pen_color+" ");
    os.print(fill_color+" ");
    os.print(depth+" ");
    os.print("0 ");
    os.print(area_fill+" ");
    os.print(style_val+" ");
    os.print("1 ");				  // direction
    os.print("0.0 ");				  // angle
  }

  private int offx(int x){
    return x+xOffset;
  }
  private int offy(int y){
    return y+yOffset;
  }


  private int colorToColor(Color c){
    if (c == Color.black)
      return cBlack;
    else if (c == Color.blue)
      return cBlue;
    else if (c == Color.green)
      return cGreen;
    else if (c == Color.cyan)
      return cCyan;
    else if (c == Color.red)
      return cRed;
    else if (c == Color.magenta)
      return cMagenta;
    else if (c == Color.yellow)
      return cYellow;
    else if (c == Color.white)
      return cWhite;

    return cDefault;
  }

  
  public void translate(int x, int y)
  {
    xOffset=x;
    yOffset=y;
  }
  
  
  public Color getColor()
  {
    if (DEBUG) diagnostic("getColor()");
    return clr;
  }

  public void setBackground(Color c)
  {
    ;
  }
  

  public void setColor(Color c)
  {
    if (DEBUG) diagnostic("setColor(" + c + ")");
    if (c != null)
      clr = c;
    pen_color = colorToColor(c);
    //    throw new RuntimeException("create() not implemented");
  }

  public void setPaintMode()
  {
    if (DEBUG) diagnostic("setPaintMode()");
    throw new RuntimeException("setPaintMode() not implemented");
  }


  public void setXORMode(Color c1)
  {
    if (DEBUG) diagnostic("setXORMode(" + c1 + ")");
    throw new RuntimeException("setXORMode() not implemented");
  }

  public Font getFont()
  {
    if (DEBUG) diagnostic("getFont()");
    return font;
  }

  public void setFont(Font f)
  {
    if (DEBUG) diagnostic("setFont(" + f + ")");

    throw new RuntimeException("setFont() not implemented");
    //  if (f != null) {
    //    this.font = f;
    //    String fontName = font.getName();
    //    int fontStyle = font.getStyle();

    //    if (fontName.startsWith("Times"))
    //      {
    //        switch (fontStyle)
    //          {
    //          case Font.PLAIN:
    //            fontName = "Times-Roman"; break;
    //          case Font.BOLD:
    //            fontName = "Times-Bold"; break;
    //          case Font.ITALIC:
    //            fontName = "Times-Italic"; break;
    //          case (Font.ITALIC + Font.BOLD):
    //            fontName += "Times-BoldItalic"; break;
    //          }
    //      }     
    //    else     
    //      {
    //        switch (fontStyle)
    //          {
    //          case Font.PLAIN:
    //            break;
    //          case Font.BOLD:
    //            fontName += "-Bold"; break;
    //          case Font.ITALIC:
    //            fontName += "-Oblique"; break;
    //          case (Font.ITALIC + Font.BOLD):
    //            fontName += "-BoldOblique"; break;
    //          }
    //      }
    //  }
  }

  public FontMetrics getFontMetrics()
  {
    if (DEBUG) diagnostic("getFontMetrics()");
    return getFontMetrics(getFont());
  }

  public FontMetrics getFontMetrics(Font f)
  {
    if (DEBUG) diagnostic("getFontMetrics(" + f + ")");
    return Toolkit.getDefaultToolkit().getFontMetrics(f);
  }


  public Rectangle getClipRect()
  {
    if (DEBUG) diagnostic("getClipRect()");
    return clippingRect;
  }

  public Shape getClip()
  {
    if (DEBUG) diagnostic("getClip()");
    return clippingRect;
  }

  public void setClip(Shape s)
  {
    if (DEBUG) diagnostic("setClip(" + s + ")");
    Rectangle r = s.getBounds();
    setClip(r.x, r.y, r.width, r.height);
  }


  public Rectangle getClipBounds()
  {
    if (DEBUG) diagnostic("getClipBounds()");
    return clippingRect;
  }

  public void setClip(int x, int y, int width, int height)
  {
    if (DEBUG) diagnostic("setClip("+ x + ", " +y + ", " +width + ", " + height +")");
    clippingRect = new Rectangle(x,y,width,height);
    throw new RuntimeException("setClip() not implemented");
  }

  public void clipRect(int x, int y, int width, int height)
  {
    if (DEBUG) diagnostic("clipRect(" + x + ", " +y + ", " +width + ", " +height +")");
    setClip(x, y, width, height);
  }

  public void copyArea(int x, int y, int width, int height, int dx, int dy)
  {
    if (DEBUG) diagnostic("copyArea(" + x + ", " +y + ", " +width + ", " +height + ", " +dx + ", " +dy +")");
    throw new RuntimeException("copyArea not supported");
  }

  public void drawLine(int x1, int y1, int x2, int y2) 
  {
    if (DEBUG) diagnostic("drawLine(" + x1 + ", " +y1 + ", " +x2 + ", " +y2 +")");
    preparePolyline(1);
    os.println(2);				  // nb of points
    os.println(offx(x1)+" "+offy(y1)+" "+offx(x2)+" "+offy(y2));
  }


  private void rectangle(int x, int y, int width, int height){
    rectangle(x,y,width,height,2);
  }
  private void rectangle(int x, int y, int width, int height,int s){
    preparePolyline(s);
    os.println(5);
    os.print(offx(x)+" "+offy(y)+" ");
    os.print(offx(x+width)+" "+offy(y)+" ");
    os.print(offx(x+width)+" "+offy(y+height)+" ");
    os.print(offx(x)+" "+offy(y+height)+" ");
    os.println(offx(x)+" "+offy(y));
  }

  public void fillRect(int x, int y, int width, int height) 
  {
    if (DEBUG) diagnostic("fillRect(" + x + ", " +y + ", " +width + ", " +height+")");
    //    doRect(x,y,width,height,true);
    int f = fill_color;
    int af = area_fill;
    fill_color=pen_color;
    area_fill = afFullSaturation;
    rectangle(x,y,width,height);
    fill_color=f;
    area_fill = af;
  }

  public void drawRect(int x, int y, int width, int height) 
  {   
    if (DEBUG) diagnostic("drawRect(" + x + ", " +y + ", " +width + ", " +height+")");
    rectangle(x,y,width,height);
  }
  
  public void clearRect(int x, int y, int width, int height) 
  {
    if (DEBUG) diagnostic("clearRect(" + x + ", " +y + ", " +width + ", " +height+")");
    int f = fill_color;
    int af = area_fill;
    int pc = pen_color;
    pen_color=fill_color=cWhite;
    area_fill = afFullSaturation;
    rectangle(x,y,width,height);
    pen_color=pc;
    fill_color=f;
    area_fill = af;
  }

  public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) 
  {
    if (DEBUG) diagnostic("drawRoundRect(" + x + ", " +y + ", " +width + ", " +height + ", " + arcWidth + ", " + arcHeight +")");
    int r = radius;
    radius = arcWidth;
    rectangle(x,y,width,height,4);
    radius = r;
  }

  public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) 
  {
    if (DEBUG) diagnostic("fillRoundRect(" + x + ", " +y + ", " +width + ", " +height + ", " +arcWidth + ", " +arcHeight+")");
    int r = radius;
    int f = fill_color;
    int af = area_fill;
    fill_color=pen_color;
    area_fill = afFullSaturation;
    radius = arcWidth;
    rectangle(x,y,width,height,4);
    radius = r;
    fill_color=f;
    area_fill = af;
  }

  public void draw3DRect(int x, int y, int width, int height, boolean raised) 
  {
    if (DEBUG) diagnostic("draw3DRect(" + x + ", " +y + ", " +width + ", " +height + ", " +raised+")");
    throw new RuntimeException("draw3DRect not supported");
  }    

  public void fill3DRect(int x, int y, int width, int height, boolean raised) 
  {
    if (DEBUG) diagnostic("fill3DRect(" + x + ", " +y + ", " +width + ", " +height + ", " +raised+")");
    throw new RuntimeException("fill3DRect not supported");
  }    

  public void drawOval(int x, int y, int width, int height) 
  {
    if (DEBUG) diagnostic("drawOval(" + x + ", " +y + ", " +width + ", " +height +")");
    //    throw new RuntimeException("copyArea not supported");
    prepareOval(1);
    os.print(offx(x)+" ");
    os.print(offy(y)+" ");
    os.print(width+" ");
    os.print(height+" ");
    os.print(offx(x)+" ");
    os.print(offy(y)+" ");
    os.print((offx(x)+width)+" ");
    os.print(offy(y)+" ");
  }

  public void fillOval(int x, int y, int width, int height) 
  {
    if (DEBUG) diagnostic("fillOval(" + x + ", " +y + ", " +width + ", " +height+")");

    int f = fill_color;
    int af = area_fill;
    fill_color=pen_color;
    area_fill = afFullSaturation;

    prepareOval(1);
    os.print(offx(x)+" ");
    os.print(offy(y)+" ");
    os.print(width+" ");
    os.print(height+" ");
    os.print(offx(x)+" ");
    os.print(offy(y)+" ");
    os.print((offx(x)+width)+" ");
    os.println(offy(y));

    fill_color=f;
    area_fill = af;
  }

  protected void doArc(int x, int y, int width, int height,
                     int startAngle, int arcAngle, boolean fill)
  {
    if (DEBUG) diagnostic("doArc(" + x + ", " +y + ", " +width + ", " +height +
                          startAngle + ", " +arcAngle + ", " +fill+")");
    throw new RuntimeException("doArc not supported");
  }


  public void drawArc(int x, int y, int width, int height,
                      int startAngle, int arcAngle) 
  {
    if (DEBUG) diagnostic("drawArc(" + x + ", " +y + ", " +width + ", " +height + ", " +startAngle + ", " +arcAngle+")");
    throw new RuntimeException("drawArc not supported");
  }

  public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle) 
  {
    if (DEBUG) diagnostic("fillArc(" + x + ", " +y + ", " +width + ", " +height + ", " +startAngle + ", " +arcAngle+")");
    throw new RuntimeException("fillArc not supported");
  }


  public void drawPolyline(int xPoints[], int yPoints[], int nPoints) 
  {
    if (DEBUG) diagnostic("drawPolyline(" + xPoints.length + ", " +yPoints.length + ", " +nPoints+")");
    preparePolyline(1);
    os.println(nPoints);			  // nb of points
    for (int i=0; i<nPoints; i++)
      os.print(offx(xPoints[i])+" "+offy(yPoints[i])+" ");
    os.println();
  }

  public void drawPolygon(int xPoints[], int yPoints[], int nPoints) 
  {
    if (DEBUG) diagnostic("drawPolygon(" + xPoints.length + ", " +yPoints.length + ", " +nPoints+")");
    throw new RuntimeException("drawPolygon not supported");
  }

  public void drawPolygon(Polygon p) 
  {
    if (DEBUG) diagnostic("drawPolygon(" + p + ")");
    throw new RuntimeException("drawPolygon not supported");
  }
  
  public void fillPolygon(int xPoints[], int yPoints[], int nPoints) 
  {
    if (DEBUG) diagnostic("fillPolygon(" + xPoints.length + ", " +yPoints.length + ", " +nPoints+")");
    throw new RuntimeException("fillPolygon not supported");
  }

  public void fillPolygon(Polygon p) 
  {
    if (DEBUG) diagnostic("fillPolygon(" + p + ")");
    throw new RuntimeException("fillPolygon not supported");
  }

  public void drawString(String str, int x, int y) 
  {
    if (DEBUG) diagnostic("drawString(" + str + ", " +x + ", " +y +")");
    
    os.print("4 ");
    os.print("0 ");				  // left justi
    os.print(pen_color+" ");
    os.print(depth+" ");
    os.print("0 ");				  // pen style
    os.print("0 ");				  // rigid font
    os.print("12 ");
    os.print("0.0 ");
    os.print("4 ");
    os.print("90 ");				  // height
    os.print((90*str.length())+" ");		  // width
    os.print(offx(x)+" "+offy(y)+" ");
    os.println(str+"\\001");
  }

  public void drawChars(char data[], int offset, int length, int x, int y) 
  {
    if (DEBUG) diagnostic("drawChars(" + data.length + ", " +offset + ", " +length + ", " +x + ", " +y +")");
    throw new RuntimeException("drawChars not supported");
  }

  public void drawBytes(byte data[], int offset, int length, int x, int y) 
  {
    if (DEBUG) diagnostic("drawBytes(" + data.length + ", " +offset + ", " +length + ", " +x + ", " +y +")");
    throw new RuntimeException("drawBytes not supported");
  }


  protected boolean doImage(Image img, int x, int y, int width, int height,
                          int sx, int sy, int sw, int sh,
                          ImageObserver observer, Color bgcolor) 
  {
    if (DEBUG) diagnostic("doImage(" + img + ", " +x + ", " +y + ", " +width + ", " + height + ", " + sx + ", " + sy + ", " + sw + ", " + sh + ", " + observer + ", " +bgcolor +")");
    
    throw new RuntimeException("doImage not supported");
  }
  
  public boolean drawImage(Image img, int x, int y,
                           ImageObserver observer) 
  {
    if (DEBUG) diagnostic("drawImage(" + img + ", " +x + ", " +y + ", " +observer+")");
    throw new RuntimeException("drawImage not supported");
  }
  
  public boolean drawImage(Image img, int x1, int y1,
                           int x2, int y2, 
                           int x3, int y3, int x4, int y4, 
                           ImageObserver observer) 
  {
    if (DEBUG) diagnostic("drawImage(" + img + ", " +x1 + ", " +y1 + ", " +x2 + ", " +y2 + ", " +x3 + ", " +y3 + ", " +x4 + ", " +y4 + ", " +observer +")"); 
    throw new RuntimeException("drawImage not supported");
  }

  public boolean drawImage(Image img, int x1, int y1,
                           int x2, int y2, 
                           int x3, int y3, int x4, int y4, 
                           Color c, 
                           ImageObserver observer) 
  {
    if (DEBUG) diagnostic("drawImage(" + img + ", " +x1 + ", " +y1 + ", " +x2 + ", " +y2 + ", " +x3 + ", " +y3 + ", " +x4 + ", " +y4 + ", " +c + ", " +observer+")");
    throw new RuntimeException("drawImage not supported");
  }

  public boolean drawImage(Image img, int x, int y,
                           int width, int height, 
                           ImageObserver observer) 
  {
    if (DEBUG) diagnostic("drawImage(" + img +"," +x + ", " +y + ", " +width + ", " +height + ", " + observer +")");
    throw new RuntimeException("drawImage not supported");
  }

  public boolean drawImage(Image img, int x, int y, Color bgcolor,
                           ImageObserver observer) 
  {
    if (DEBUG) diagnostic("drawImage("+ img + ", " +x + ", " +y + ", " +bgcolor + ", " + observer +")");
    throw new RuntimeException("drawImage not supported");
  }

  public boolean drawImage(Image img, int x, int y,
                           int width, int height, Color bgcolor,
                           ImageObserver observer) 
  {
    if (DEBUG) diagnostic("drawImage(" + img + ", " +x + ", " +y + ", " +width + ", " +height + ", " +bgcolor + ", " +observer +")");
    throw new RuntimeException("drawImage not supported");
  }
  
  public void dispose() 
  {
    if (DEBUG) diagnostic("dispose() ");
    throw new RuntimeException("dispose not supported");
  }

  public void finalize() 
  {
    if (DEBUG) diagnostic("finalize() ");
    throw new RuntimeException("finalize not supported");
  }

  public String toString() 
  {
    if (DEBUG) diagnostic("toString() ");
    return getClass().getName() + "[font=" + getFont() + ",color=" + getColor() + "]";
  }

  protected void emitProlog() 
  {
    os.println("#FIG 3.2");
    os.println("Landscape");
    os.println("Center");
    os.println("Metric");
    os.println("A4");
    os.println("100.0");			  // magnification
    os.println("Single");
    os.println("-2");				  // transparent color -2:none
    os.println("1200 2");			  // Resolution + origin
  }

  protected void diagnostic(String s)
  {
    os.print("#  ");
    os.print(hashCode());
    os.print(": ");
    os.println(s);
  }

}


  



